/* Generated by the protocol buffer compiler.  DO NOT EDIT! */

/* Do not generate deprecated warnings for self */
#ifndef PROTOBUF_C_NO_DEPRECATED
#define PROTOBUF_C_NO_DEPRECATED
#endif

#include "pinba.pb-c.h"
void   pinba__request__init
                     (Pinba__Request         *message)
{
  static Pinba__Request init_value = PINBA__REQUEST__INIT;
  *message = init_value;
}
size_t pinba__request__get_packed_size
                     (const Pinba__Request *message)
{
  PROTOBUF_C_ASSERT (message->base.descriptor == &pinba__request__descriptor);
  return protobuf_c_message_get_packed_size ((const ProtobufCMessage*)(message));
}
size_t pinba__request__pack
                     (const Pinba__Request *message,
                      uint8_t       *out)
{
  PROTOBUF_C_ASSERT (message->base.descriptor == &pinba__request__descriptor);
  return protobuf_c_message_pack ((const ProtobufCMessage*)message, out);
}
size_t pinba__request__pack_to_buffer
                     (const Pinba__Request *message,
                      ProtobufCBuffer *buffer)
{
  PROTOBUF_C_ASSERT (message->base.descriptor == &pinba__request__descriptor);
  return protobuf_c_message_pack_to_buffer ((const ProtobufCMessage*)message, buffer);
}
Pinba__Request *
       pinba__request__unpack
                     (ProtobufCAllocator  *allocator,
                      size_t               len,
                      const uint8_t       *data)
{
  return (Pinba__Request *)
     protobuf_c_message_unpack (&pinba__request__descriptor,
                                allocator, len, data);
}
void   pinba__request__free_unpacked
                     (Pinba__Request *message,
                      ProtobufCAllocator *allocator)
{
  PROTOBUF_C_ASSERT (message->base.descriptor == &pinba__request__descriptor);
  protobuf_c_message_free_unpacked ((ProtobufCMessage*)message, allocator);
}
static const ProtobufCFieldDescriptor pinba__request__field_descriptors[23] =
{
  {
    .name              = "hostname",
    .id                = 1,
    .label             = PROTOBUF_C_LABEL_REQUIRED,
    .type              = PROTOBUF_C_TYPE_STRING,
    .quantifier_offset = 0,
    .offset            = PROTOBUF_C_OFFSETOF(Pinba__Request, hostname),
    .descriptor        = NULL,
    .default_value     = NULL,
  },
  {
    .name              = "server_name",
    .id                = 2,
    .label             = PROTOBUF_C_LABEL_REQUIRED,
    .type              = PROTOBUF_C_TYPE_STRING,
    .quantifier_offset = 0,
    .offset            = PROTOBUF_C_OFFSETOF(Pinba__Request, server_name),
    .descriptor        = NULL,
    .default_value     = NULL,
  },
  {
    .name              = "script_name",
    .id                = 3,
    .label             = PROTOBUF_C_LABEL_REQUIRED,
    .type              = PROTOBUF_C_TYPE_STRING,
    .quantifier_offset = 0,
    .offset            = PROTOBUF_C_OFFSETOF(Pinba__Request, script_name),
    .descriptor        = NULL,
    .default_value     = NULL,
  },
  {
    .name              = "request_count",
    .id                = 4,
    .label             = PROTOBUF_C_LABEL_REQUIRED,
    .type              = PROTOBUF_C_TYPE_UINT32,
    .quantifier_offset = 0,
    .offset            = PROTOBUF_C_OFFSETOF(Pinba__Request, request_count),
    .descriptor        = NULL,
    .default_value     = NULL,
  },
  {
    .name              = "document_size",
    .id                = 5,
    .label             = PROTOBUF_C_LABEL_REQUIRED,
    .type              = PROTOBUF_C_TYPE_UINT32,
    .quantifier_offset = 0,
    .offset            = PROTOBUF_C_OFFSETOF(Pinba__Request, document_size),
    .descriptor        = NULL,
    .default_value     = NULL,
  },
  {
    .name              = "memory_peak",
    .id                = 6,
    .label             = PROTOBUF_C_LABEL_REQUIRED,
    .type              = PROTOBUF_C_TYPE_UINT32,
    .quantifier_offset = 0,
    .offset            = PROTOBUF_C_OFFSETOF(Pinba__Request, memory_peak),
    .descriptor        = NULL,
    .default_value     = NULL,
  },
  {
    .name              = "request_time",
    .id                = 7,
    .label             = PROTOBUF_C_LABEL_REQUIRED,
    .type              = PROTOBUF_C_TYPE_FLOAT,
    .quantifier_offset = 0,
    .offset            = PROTOBUF_C_OFFSETOF(Pinba__Request, request_time),
    .descriptor        = NULL,
    .default_value     = NULL,
  },
  {
    .name              = "ru_utime",
    .id                = 8,
    .label             = PROTOBUF_C_LABEL_REQUIRED,
    .type              = PROTOBUF_C_TYPE_FLOAT,
    .quantifier_offset = 0,
    .offset            = PROTOBUF_C_OFFSETOF(Pinba__Request, ru_utime),
    .descriptor        = NULL,
    .default_value     = NULL,
  },
  {
    .name              = "ru_stime",
    .id                = 9,
    .label             = PROTOBUF_C_LABEL_REQUIRED,
    .type              = PROTOBUF_C_TYPE_FLOAT,
    .quantifier_offset = 0,
    .offset            = PROTOBUF_C_OFFSETOF(Pinba__Request, ru_stime),
    .descriptor        = NULL,
    .default_value     = NULL,
  },
  {
    .name              = "timer_hit_count",
    .id                = 10,
    .label             = PROTOBUF_C_LABEL_REPEATED,
    .type              = PROTOBUF_C_TYPE_UINT32,
    .quantifier_offset = PROTOBUF_C_OFFSETOF(Pinba__Request, n_timer_hit_count),
    .offset            = PROTOBUF_C_OFFSETOF(Pinba__Request, timer_hit_count),
    .descriptor        = NULL,
    .default_value     = NULL,
  },
  {
    .name              = "timer_value",
    .id                = 11,
    .label             = PROTOBUF_C_LABEL_REPEATED,
    .type              = PROTOBUF_C_TYPE_FLOAT,
    .quantifier_offset = PROTOBUF_C_OFFSETOF(Pinba__Request, n_timer_value),
    .offset            = PROTOBUF_C_OFFSETOF(Pinba__Request, timer_value),
    .descriptor        = NULL,
    .default_value     = NULL,
  },
  {
    .name              = "timer_tag_count",
    .id                = 12,
    .label             = PROTOBUF_C_LABEL_REPEATED,
    .type              = PROTOBUF_C_TYPE_UINT32,
    .quantifier_offset = PROTOBUF_C_OFFSETOF(Pinba__Request, n_timer_tag_count),
    .offset            = PROTOBUF_C_OFFSETOF(Pinba__Request, timer_tag_count),
    .descriptor        = NULL,
    .default_value     = NULL,
  },
  {
    .name              = "timer_tag_name",
    .id                = 13,
    .label             = PROTOBUF_C_LABEL_REPEATED,
    .type              = PROTOBUF_C_TYPE_UINT32,
    .quantifier_offset = PROTOBUF_C_OFFSETOF(Pinba__Request, n_timer_tag_name),
    .offset            = PROTOBUF_C_OFFSETOF(Pinba__Request, timer_tag_name),
    .descriptor        = NULL,
    .default_value     = NULL,
  },
  {
    .name              = "timer_tag_value",
    .id                = 14,
    .label             = PROTOBUF_C_LABEL_REPEATED,
    .type              = PROTOBUF_C_TYPE_UINT32,
    .quantifier_offset = PROTOBUF_C_OFFSETOF(Pinba__Request, n_timer_tag_value),
    .offset            = PROTOBUF_C_OFFSETOF(Pinba__Request, timer_tag_value),
    .descriptor        = NULL,
    .default_value     = NULL,
  },
  {
    .name              = "dictionary",
    .id                = 15,
    .label             = PROTOBUF_C_LABEL_REPEATED,
    .type              = PROTOBUF_C_TYPE_STRING,
    .quantifier_offset = PROTOBUF_C_OFFSETOF(Pinba__Request, n_dictionary),
    .offset            = PROTOBUF_C_OFFSETOF(Pinba__Request, dictionary),
    .descriptor        = NULL,
    .default_value     = NULL,
  },
  {
    .name              = "status",
    .id                = 16,
    .label             = PROTOBUF_C_LABEL_OPTIONAL,
    .type              = PROTOBUF_C_TYPE_UINT32,
    .quantifier_offset = PROTOBUF_C_OFFSETOF(Pinba__Request, has_status),
    .offset            = PROTOBUF_C_OFFSETOF(Pinba__Request, status),
    .descriptor        = NULL,
    .default_value     = NULL,
  },
  {
    .name              = "memory_footprint",
    .id                = 17,
    .label             = PROTOBUF_C_LABEL_OPTIONAL,
    .type              = PROTOBUF_C_TYPE_UINT32,
    .quantifier_offset = PROTOBUF_C_OFFSETOF(Pinba__Request, has_memory_footprint),
    .offset            = PROTOBUF_C_OFFSETOF(Pinba__Request, memory_footprint),
    .descriptor        = NULL,
    .default_value     = NULL,
  },
  {
    .name              = "requests",
    .id                = 18,
    .label             = PROTOBUF_C_LABEL_REPEATED,
    .type              = PROTOBUF_C_TYPE_MESSAGE,
    .quantifier_offset = PROTOBUF_C_OFFSETOF(Pinba__Request, n_requests),
    .offset            = PROTOBUF_C_OFFSETOF(Pinba__Request, requests),
    .descriptor        = &pinba__request__descriptor,
    .default_value     = NULL,
  },
  {
    .name              = "schema",
    .id                = 19,
    .label             = PROTOBUF_C_LABEL_OPTIONAL,
    .type              = PROTOBUF_C_TYPE_STRING,
    .quantifier_offset = 0,
    .offset            = PROTOBUF_C_OFFSETOF(Pinba__Request, schema),
    .descriptor        = NULL,
    .default_value     = NULL,
  },
  {
    .name              = "tag_name",
    .id                = 20,
    .label             = PROTOBUF_C_LABEL_REPEATED,
    .type              = PROTOBUF_C_TYPE_UINT32,
    .quantifier_offset = PROTOBUF_C_OFFSETOF(Pinba__Request, n_tag_name),
    .offset            = PROTOBUF_C_OFFSETOF(Pinba__Request, tag_name),
    .descriptor        = NULL,
    .default_value     = NULL,
  },
  {
    .name              = "tag_value",
    .id                = 21,
    .label             = PROTOBUF_C_LABEL_REPEATED,
    .type              = PROTOBUF_C_TYPE_UINT32,
    .quantifier_offset = PROTOBUF_C_OFFSETOF(Pinba__Request, n_tag_value),
    .offset            = PROTOBUF_C_OFFSETOF(Pinba__Request, tag_value),
    .descriptor        = NULL,
    .default_value     = NULL,
  },
  {
    .name              = "timer_ru_utime",
    .id                = 22,
    .label             = PROTOBUF_C_LABEL_REPEATED,
    .type              = PROTOBUF_C_TYPE_FLOAT,
    .quantifier_offset = PROTOBUF_C_OFFSETOF(Pinba__Request, n_timer_ru_utime),
    .offset            = PROTOBUF_C_OFFSETOF(Pinba__Request, timer_ru_utime),
    .descriptor        = NULL,
    .default_value     = NULL,
  },
  {
    .name              = "timer_ru_stime",
    .id                = 23,
    .label             = PROTOBUF_C_LABEL_REPEATED,
    .type              = PROTOBUF_C_TYPE_FLOAT,
    .quantifier_offset = PROTOBUF_C_OFFSETOF(Pinba__Request, n_timer_ru_stime),
    .offset            = PROTOBUF_C_OFFSETOF(Pinba__Request, timer_ru_stime),
    .descriptor        = NULL,
    .default_value     = NULL,
  },
};
static const unsigned pinba__request__field_indices_by_name[] = {
  14,   /* field[14] = dictionary */
  4,   /* field[4] = document_size */
  0,   /* field[0] = hostname */
  16,   /* field[16] = memory_footprint */
  5,   /* field[5] = memory_peak */
  3,   /* field[3] = request_count */
  6,   /* field[6] = request_time */
  17,   /* field[17] = requests */
  8,   /* field[8] = ru_stime */
  7,   /* field[7] = ru_utime */
  18,   /* field[18] = schema */
  2,   /* field[2] = script_name */
  1,   /* field[1] = server_name */
  15,   /* field[15] = status */
  19,   /* field[19] = tag_name */
  20,   /* field[20] = tag_value */
  9,   /* field[9] = timer_hit_count */
  22,   /* field[22] = timer_ru_stime */
  21,   /* field[21] = timer_ru_utime */
  11,   /* field[11] = timer_tag_count */
  12,   /* field[12] = timer_tag_name */
  13,   /* field[13] = timer_tag_value */
  10,   /* field[10] = timer_value */
};
static const ProtobufCIntRange pinba__request__number_ranges[1 + 1] =
{
  { 1, 0 },
  { 0, 23 }
};
const ProtobufCMessageDescriptor pinba__request__descriptor =
{
  .magic                 = PROTOBUF_C_MESSAGE_DESCRIPTOR_MAGIC,
  .name                  = "Pinba.Request",
  .short_name            = "Request",
  .c_name                = "Pinba__Request",
  .package_name          = "Pinba",
  .sizeof_message        = sizeof(Pinba__Request),
  .n_fields              = 23,
  .fields                = pinba__request__field_descriptors,
  .fields_sorted_by_name = pinba__request__field_indices_by_name,
  .n_field_ranges        = 1,
  .field_ranges          = pinba__request__number_ranges,
  .message_init          = (ProtobufCMessageInit) pinba__request__init,
};
